/* 
CM2306: Communication Networks - Group U28, 2023
Smart Plant Care System
Hamza Khan - c1981711
Finlay Houlton - c21026424 
Kaviya Sundarapalan - c21053687
Jake Samuels - c21068235

Sketch uses 25584 bytes (89%) of program storage space. Maximum is 28672 bytes.
Global variables use 1795 bytes (70%) of dynamic memory, leaving 765 bytes for local variables. Maximum is 2560 bytes.
*/
// SOFTWARE LIBRARIES
#include <rgb_lcd.h>  // LCD screen
#include <Wire.h>     // Support for I2C communication
#include <DHT.h>      // Temp & Humidity sensor
#include <Servo.h>    // Servo motor

#define DHTPIN 6      // Define DHT sensor pin and model
#define DHTTYPE DHT11


// ### CREATE OBJECTS ### //

rgb_lcd lcd;                // LCD screen
Servo myservo;              // Servo motor
DHT dht(DHTPIN, DHTTYPE);   // DHT sensor

// ### SYSTEM VARIABLES ### //

const int buttonPin = 4;  // the number of the pushbutton pin, D4
const int lightPin = A0;  // Light sensor pin
const int buzzerPin = 8;  // Buzzer's pin
int buttonState;          // the state of the button

unsigned long lastPressed;  // Monitors when plant was last watered

float temp; float hum; int light; float lightIntensity; // Variables for sensor data
float minSafeTemp = 13.0; float maxSafeTemp = 32.0;
float minSafeHum = 35.0;  float maxSafeHum = 100.0;
String watering = "";   // For animating the "WATERING ..." message
String soundPlayed = "no";  // Ensures buzzer only sounds when the screen first turns red, not constantly
char recvChar; String WholeCommand = "";  // For BLE interface
unsigned long waterInterval = 90000;   // The interval in ms that the plant is watered. For demo purposes, default = 90 secs (1.5 min, 90,000 ms)

void setup() {
  // setup code here, runs once
  pinMode(buttonPin, INPUT);    // Set button and buzzer pin modes
  pinMode(buzzerPin, OUTPUT);
 
  myservo.attach(5);  // Set servo port
  dht.begin();        // Initialise DHT sensor's object
  lcd.begin(16, 2);   // LCD has 16 rows and 2 columns
  lcd.setRGB(255, 255, 255);  // Set the LCD screen to white

  Serial.begin(9600);       // Begin serial monitor at baud rate 9600
  setupBluetoothConnection(); // Initialise BLE module

}

// ### FUNCTIONS ### //
void setupBluetoothConnection() { // Adapted from LAB 10
  Serial1.begin(9600);
  Serial1.println("AT");  // Use AT commands
  delay(400);
  readAndDiscardResponse();
  Serial1.println("AT+RESET"); // reset the module
  readAndDiscardResponse();
  Serial1.println("AT+ROLE0"); // set the role as peripheral.
  delay(400);
  readAndDiscardResponse();
  Serial1.println("AT+NAMEGroupU28"); // set the name as GroupU28
  readAndDiscardResponse();
  Serial1.println("AT+CON?");  // send an AT command to get the connection status
  Serial1.flush();
}
void readAndDiscardResponse() { // For when Serial1 response doesn't matter
  delay(100);
  while (Serial1.available()) {
    Serial1.read();
  }
}
String SerialString()   // For reading Serial communication
{
  String inputString = "";
  while (Serial.available()){
    char inputChar = (char)Serial.read();
    inputString += inputChar;
  }
  return inputString;
}

String readInput() {    // Takes input from connected BLE device
  String receivedCommand = "";
  delay(100);
  while(Serial1.available()) {
  recvChar = Serial1.read();
  if (recvChar == '\n') {   // Reads character by character and ends at newline
    if (receivedCommand.startsWith("OK+")) {  // If a device disconnects and connects, filter out AT+CON responses from their message
      receivedCommand = receivedCommand.substring(receivedCommand.lastIndexOf("CONN")+4, receivedCommand.length()-1);
    }
    else {
      receivedCommand = receivedCommand.substring(0, receivedCommand.length()-1);
    }
  }
  else {
    receivedCommand += recvChar;  // Append character to current message
  }
  }
  return receivedCommand;
}

void readSensors(){ // Take a reading from the DHT and light sensors, and check the button state
  buttonState = digitalRead(buttonPin);
  temp = dht.readTemperature();
  hum = dht.readHumidity();
  light = analogRead(lightPin);
  lightIntensity = (float)light / 760.0 * 100.0;  // convert light to a percentage of the max reading
}

void checkDanger(){ // Check if sensor values are out of range, turn red and play sound (once) if true
  if((temp < minSafeTemp) || (temp > maxSafeTemp) || (hum < minSafeHum) || (hum > maxSafeHum)){
    delay(2500);
    temp = dht.readTemperature();
    hum = dht.readHumidity();
    if((temp < minSafeTemp) || (temp > maxSafeTemp) || (hum < minSafeHum) || (hum > maxSafeHum)) {
    // make red
    lcd.setRGB(255, 5, 5);

    if (soundPlayed.equals("no")){
      // sound buzzer
      tone(buzzerPin, 800);
      delay(200);
      tone(buzzerPin, 1000);
      delay(200);
      tone(buzzerPin, 1200);
      delay(200);
      noTone(buzzerPin);
      soundPlayed = "yes";
    }
    lcd.clear();
    delay(100);
    if(temp < minSafeTemp){
      lcd.setCursor(0, 0);
      lcd.print("TEMPERATURE");
      lcd.setCursor(0, 1);
      lcd.print("TOO LOW");
    }
    else if (temp > maxSafeTemp) {
      lcd.setCursor(0, 0);
      lcd.print("TEMPERATURE");
      lcd.setCursor(0, 1);
      lcd.print("TOO HIGH");
    }
    else if (hum < minSafeHum) {
      lcd.setCursor(0, 0);
      lcd.print("HUMIDITY");
      lcd.setCursor(0, 1);
      lcd.print("TOO LOW");
    }
    else {
      lcd.setCursor(0, 0);
      lcd.print("HUMIDITY");
      lcd.setCursor(0, 1);
      lcd.print("TOO HIGH");
    }
    delay(2000);
    }
  }
  else {    // Reset LCD to white
    // make white
    lcd.setRGB(255, 255, 255);
    soundPlayed = "no"; // Allow buzzer to go off again next time the screen is red
  }
}

String findTimeDiff() { // Find the time since the plant was last watered, returns string as __h __m
  int difference = ((millis()/1000) - lastPressed);
  int hours = difference / 3600;
  int minutes = (difference % 3600) / 60;
  String timeDiff = ""+String(hours)+"h "+String(minutes)+"m";
  return timeDiff;
}

void writeSensorData() {  // Write the sensor data to the LCD screen
  lcd.clear();

  // Write temp and humidity
  lcd.setCursor(0, 0);
  lcd.print("T:"+String(temp, 1)+"C");
  lcd.setCursor(0, 1);
  lcd.print("H:"+String(hum, 0)+"%");

  // Write time since last water
  lcd.setCursor(8, 0);
  
  String timeDiff = findTimeDiff();
  lcd.print("W:"+timeDiff);

  // Write light intensity
  lcd.setCursor(8, 1);
  lcd.print("L:"+String(lightIntensity)+"%");
}

void buttonDown() {   // When button is pressed,
  myservo.write(0);  // tell servo to go to that position (represents opening solenoid valve and allowing water to flow)
    delay(15);         // waits 15ms for the servo to reach the position
      
    while (buttonState == LOW) {  // Keep watering until button is released
      lcd.clear();  lcd.setCursor(3, 0); lcd.print("WATERING");
                    lcd.setCursor(5, 1); lcd.print(watering); // Animates LCD message
      if (watering.equals("")) watering = "."; else if (watering.equals(".")) watering = ".."; else if (watering.equals("..")) watering = "..."; else if (watering.equals("...")) watering = "";
      buttonState = digitalRead(buttonPin);
      delay(250);
    }
    lastPressed = millis()/1000;  // Update that the plant was watered
    myservo.write(90);    // Represents closing the solenoid valve, stopping water
}

void waterPlant() {   // The same as button down but without needing the button
  myservo.write(0);
  float counter = 0.00;
  while(counter < 5) {
    lcd.clear();  lcd.setCursor(3, 0); lcd.print("WATERING");
                    lcd.setCursor(5, 1); lcd.print(watering);
    if (watering.equals("")) watering = "."; else if (watering.equals(".")) watering = ".."; else if (watering.equals("..")) watering = "..."; else if (watering.equals("...")) watering = "";
    delay(250);
    counter += 0.250;
  }
  lastPressed = millis()/1000;
  myservo.write(90);
}

void printMenu(){ // Welcome message to BLE device
  Serial1.println("Welcome! Enter a number to access settings.");
  Serial1.println("1 - Temperature");
  Serial1.println("2 - Humidity");
  Serial1.println("3 - Watering");
  Serial1.println("4 - Current Info");
}
void tempMenu() {   // Menu for adjusting temperature settings from BLE
  String receivedCommand = "";
  Serial1.println("");
  Serial1.println("Adjusting temperature settings.");
  Serial1.println("Current temperature: "+String(temp)+"C");
  Serial1.println("Min: "+String(minSafeTemp)+"C");
  Serial1.println("Max: "+String(maxSafeTemp)+"C");
  Serial1.println("==========");
  Serial1.println("1 - Change min safe temp");
  Serial1.println("2 - Change max safe temp");
  Serial1.println("3 - Cancel");

  while (receivedCommand.equals("")) { receivedCommand = readInput(); }
  if (receivedCommand.equals("1")) {
    Serial1.println("");
    Serial1.println("Enter new minimum:");
    receivedCommand = "";
    while (receivedCommand.equals("")) { receivedCommand = readInput(); }
    if ((receivedCommand.toFloat()) > maxSafeTemp) {    // Don't allow min to be greater than max
      while((receivedCommand.toFloat()) > maxSafeTemp) {
        Serial1.println("ERROR: New minimum cannot be greater than current maximum.");
        receivedCommand = ""; while (receivedCommand.equals("")) { receivedCommand = readInput(); }
      }
    }
    minSafeTemp = receivedCommand.toFloat();
    Serial1.println("");
    Serial1.println("Updated: "+String(minSafeTemp));
    return;
  }
  else if (receivedCommand.equals("2")) {
    receivedCommand = "";
    Serial1.println("");
    Serial1.println("Enter new maximum:");
    while (receivedCommand.equals("")) { receivedCommand = readInput(); }
    if ((receivedCommand.toFloat()) < minSafeTemp) {    // Don't allow max to be less than min
      while((receivedCommand.toFloat()) < minSafeTemp) {
        Serial1.println("ERROR: New maximum cannot be less than current minimum.");
        receivedCommand = ""; while (receivedCommand.equals("")) { receivedCommand = readInput(); }
      }
    }
    maxSafeTemp = receivedCommand.toFloat();
    Serial1.println("");
    Serial1.println("Updated: "+String(maxSafeTemp));
    return;
  }
  else {
    Serial1.println("Exiting...");
    return;
  }
}

void humMenu() {        // Menu for adjusting humidity settings from BLE
  String receivedCommand = "";
  Serial1.println("");
  Serial1.println("Adjusting humidity settings.");
  Serial1.println("Current humidity: "+String(hum)+"%");
  Serial1.println("Min: "+String(minSafeHum)+"%");
  Serial1.println("Max: "+String(maxSafeHum)+"%");
  Serial1.println("==========");
  Serial1.println("1 - Change min safe hum");
  Serial1.println("2 - Change max safe hum");
  Serial1.println("3 - Cancel");

  while (receivedCommand.equals("")) { receivedCommand = readInput(); }
  if (receivedCommand.equals("1")) {
    Serial1.println("");
    Serial1.println("Enter new minimum:");
    receivedCommand = "";
    while (receivedCommand.equals("")) { receivedCommand = readInput(); }
    if ((receivedCommand.toFloat()) > maxSafeHum) {   // Don't allow min to be greater than max
      while((receivedCommand.toFloat()) > maxSafeHum) {
        Serial1.println("ERROR: New minimum cannot be greater than current maximum.");
        receivedCommand = ""; while (receivedCommand.equals("")) { receivedCommand = readInput(); }
      }
    }
    minSafeHum = receivedCommand.toFloat();
    Serial1.println("");
    Serial1.println("Updated: "+String(minSafeHum));
    return;
  }
  else if (receivedCommand.equals("2")) {
    receivedCommand = "";
    Serial1.println("");
    Serial1.println("Enter new maximum:");
    while (receivedCommand.equals("")) { receivedCommand = readInput(); }
    if ((receivedCommand.toFloat()) < minSafeHum) {   // Don't allow max to be less than min
      while((receivedCommand.toFloat()) < minSafeHum) {
        Serial1.println("ERROR: New maximum cannot be less than current minimum.");
        receivedCommand = ""; while (receivedCommand.equals("")) { receivedCommand = readInput(); }
      }
    }
    maxSafeHum = receivedCommand.toFloat();
    Serial1.println("");
    Serial1.println("Updated: "+String(maxSafeHum));
    return;
  }
  else {
    Serial1.println("Exiting...");
    return;
  }
}
void waterMenu() {      // Menu for adjusting water settings from BLE
  String receivedCommand = "";
  String timeDiff = findTimeDiff();
  Serial1.println("");
  Serial1.println("Adjusting watering settings.");
  Serial1.println("Current interval: every "+String(waterInterval/1000)+" seconds");
  Serial1.println("Last watered: "+timeDiff);
  Serial1.println("==========");
  Serial1.println("1 - Change interval");
  Serial1.println("2 - Water now");
  Serial1.println("2 - Cancel");

  while (receivedCommand.equals("")) { receivedCommand = readInput(); }
  if (receivedCommand.equals("1")) {
    Serial1.println("");
    Serial1.println("Enter new interval (seconds):");
    receivedCommand = "";
    while (receivedCommand.equals("")) { receivedCommand = readInput(); }
    if ((receivedCommand.toFloat()) < 0) {    // Don't allow number to be 0 or less
      while((receivedCommand.toFloat()) < 0) {
        Serial1.println("ERROR: New interval must be greater than 0.");
        receivedCommand = ""; while (receivedCommand.equals("")) { receivedCommand = readInput(); }
      }
    }
    waterInterval = (receivedCommand.toFloat())*1000;
    Serial1.println("");
    Serial1.println("Updated: "+String(waterInterval/1000));
    return;
  }
  else if (receivedCommand.equals("2")) {   // Water the plant now, command from BLE
    Serial1.println("");
    Serial1.println("Watering...");
    receivedCommand = "";
    waterPlant();
    Serial1.println("Done!");
    return;
    }
  else {
    Serial1.println("Exiting...");
    receivedCommand = "";
    return;
  }
}

void getCurrentInfo() { // Read sensor info and write to BLE console
  String receivedCommand = "";
  readSensors();
  Serial1.print("=====\n Temperature: "+String(temp)+"C\n Humidity: "+String(hum)+"%\n Light: "+String(lightIntensity)+"%\n Last watered: "+findTimeDiff()+"\n=====\n");
  return;
}

void mainFlow() {   // The main flow of the sensors and LCD screen
  readSensors();    // Read the sensor data
  checkDanger();    // Check if it is outside of the safe values and act if anything is
  writeSensorData();// Write it to the screen
  if(buttonState == LOW){ // If button is pressed,
    buttonDown();         // water the plant
  }
  else {
    delay(100);
  }
  if ((millis()/1000) - lastPressed > (waterInterval/1000)) {
    waterPlant(); // Automatically water the plant if the time interval has elapsed
  }
}


void loop() {
  // main code here, runs repeatedly
  mainFlow(); // Execute main behaviour

  while(Serial1.available()){  // Receive from client
    String receivedCommand = "";  // Allows us to be controlled by a connected BLE device
    lcd.setRGB(255, 5, 255); 
    while (receivedCommand.equals("")) { receivedCommand = readInput(); } // Waits for user to send input, turns screen dark pink while waiting for input
    // Serial1.println("Received message"); // For debugging, output received BLE command
    // Serial1.println(receivedCommand);
    printMenu();
    receivedCommand = "";
    while (receivedCommand.equals("")) { receivedCommand = readInput(); } // Execute behaviour based on user choice
    if      (receivedCommand.equals("1")) { tempMenu();   receivedCommand = "-1"; break; }
    else if (receivedCommand.equals("2")) { humMenu();    receivedCommand = "-1"; break; }
    else if (receivedCommand.equals("3")) { waterMenu();  receivedCommand = "-1"; break; }
    else if (receivedCommand.equals("4")) { getCurrentInfo(); receivedCommand = "-1"; break; }
    lcd.setRGB(5, 255, 5);  // Flash screen green to show command was received
    delay(50);
  }
  while(Serial.available()){  // Send to client
    WholeCommand = SerialString();    // Gives ability to write to Serial, not really used in this project.
    Serial1.println(WholeCommand+"\n");
    lcd.setRGB(5, 5, 255);  // Flashes screen dark blue
    delay(400);
  }
  delay(100);

} // end of loop
